/*
 * filter.c
 *
 * functions (or in-line code) for all cascading filters
 *
 */

/*
 * Copyright (c) 2001 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#ifdef	TIME_WITH_SYS_TIME
#include <time.h>
#endif

#include "ourmon.h"
#include  "filter.h"

extern struct fixed_ipproto 	fixed_ipproto;
extern struct fixed_tcp3 	fixed_tcp3;

void fixed_ipproto_filter(unsigned int , struct t_entry *);
void fixed_tcp3_filter(unsigned int, struct t_entry *);
void fixed_iprange1_filter(unsigned int, struct t_entry *);
void fixed_cast_filter(unsigned int, struct t_entry *);
void topn_ip_filter(unsigned int, struct t_entry *);

ip_filter(unsigned int caplength, struct t_entry *te) 
{

	if (debug)  {
		printf("ip_filter pkt_len %d\n", caplength);
	}

	/* cascading filters - check them all ...
	*/
	if (fixed_ipproto.configured) {
		fixed_ipproto_filter(caplength, te);
	}

	if (fixed_tcp3.configured) {
		fixed_tcp3_filter(caplength, te);

	}

	if (fixed_iprange1.configured) {
		fixed_iprange1_filter(caplength, te);
	}

	if (fixed_cast.configured) {
		fixed_cast_filter(caplength, te);
	}

	/* topn manipulates linked lists and free lists
	 * block ALARM whilst in there
	 */
	block_signal();
	if (topn_ip.configured) {
		topn_ip_filter(caplength, te);
	}
	unblock_signal();
}

/* note: we do not count "non-ip" except in the case
 * of fixed_ipproto, specified ip_prpto, and fixed_cast.
 * The other filters are only for ip addresses in terms
 * of their "xtra" bytes.
 */
non_ip_filter(caplength) 
{
	if (fixed_ipproto.configured) {
		fixed_ipproto.xtra_bytes += caplength;
	}
	if (spec_ipproto.configured) {
		spec_ipproto.xtra_bytes += caplength;
	}
	if (fixed_cast.configured) {
		fixed_cast.xtra_bytes += caplength;
	}
}

void
fixed_ipproto_filter(unsigned int caplength, struct t_entry *te) 
{
	switch(te->proto) {
	case IPPROTO_TCP:
		fixed_ipproto.tcp_bytes += caplength;
		break;
	case IPPROTO_UDP:
		fixed_ipproto.udp_bytes += caplength;
		break;
	case IPPROTO_ICMP:
		fixed_ipproto.icmp_bytes += caplength;
		break;
	/* TBD. this should be a global counter
	*/
	default:
		fixed_ipproto.xtra_bytes += caplength;
		break;
	}		
}

void fixed_tcp3_filter(unsigned int caplength, struct t_entry *te) 
{
	if (te->proto == IPPROTO_TCP) {
		unsigned int dport = te->dport;
		unsigned int sport = te->sport;
		if (debug) {
			printf("tcp3, dport %d, sport %d\n", 
				dport, sport);
		}
		if (dport == fixed_tcp3.tcp_port1
			|| sport == fixed_tcp3.tcp_port1) {
			if (debug) {
				printf("tcp3, port1 b=%d\n", caplength);
			}
			fixed_tcp3.port1_bytes += caplength;
		}
		else if (dport == fixed_tcp3.tcp_port2
			|| sport == fixed_tcp3.tcp_port2) {
			if (debug) {
				printf("tcp3, port2 b=%d\n", caplength);
			}
			fixed_tcp3.port2_bytes += caplength;
		}
		else {
			if (debug) {
				printf("tcp3(p), xtra b=%d\n", caplength);
			}
			fixed_tcp3.xtra_bytes += caplength;
		}
	}
	else {
		if (debug) {
			printf("tcp3, xtra b=%d\n", caplength);
		}
		fixed_tcp3.xtra_bytes += caplength;
	}
}

/* take the src and dst in turn, and see if they belong
 * in the range.  reverse and check, else fail.
 */
void fixed_iprange1_filter(unsigned int caplength, struct t_entry *te) 
{
	register struct in_addr src1, dst1, src2, dst2;

	src1.s_addr = te->src.s_addr & fixed_iprange1.netmask1.s_addr;
	dst1.s_addr = te->dst.s_addr & fixed_iprange1.netmask2.s_addr;
	if ( src1.s_addr == fixed_iprange1.ipnet1.s_addr 
		&&
	     dst1.s_addr == fixed_iprange1.ipnet2.s_addr ) {
		
		fixed_iprange1.range1_bytes += caplength;
		return;
	}
	src2.s_addr = te->src.s_addr & fixed_iprange1.netmask2.s_addr;
	dst2.s_addr = te->dst.s_addr & fixed_iprange1.netmask1.s_addr;
	if ( src2.s_addr == fixed_iprange1.ipnet2.s_addr 
		&&
	     dst2.s_addr == fixed_iprange1.ipnet1.s_addr ) {
		fixed_iprange1.range1_bytes += caplength;
		return;
	}
	fixed_iprange1.extra_bytes += caplength;
}

/* 
 * if we are handed ethernet packets, we function at ethernet
 * level, else we try and fake it with IP level packets.
 * sort IP addresses by address dst type, multicast, broadcast, extra
 * note: this will miss non-ip mcast/bcast
 */
void fixed_cast_filter(unsigned int caplength, struct t_entry *te) 
{
	register struct in_addr dst;

	if (te->eflag) {
		if (te->bflag) {
			fixed_cast.bcast_bytes += caplength;
			return;
		}
		if (te->mflag) {
			fixed_cast.mcast_bytes += caplength;
			return;
		}
		fixed_cast.ucast_bytes += caplength;
		return;
	}
	/* it wasn't an ethernet packet
	*/
	dst.s_addr = ntohl(te->dst.s_addr);

	/* multicast IP
	*/
	if (IN_MULTICAST(dst.s_addr)) {
		fixed_cast.mcast_bytes += caplength;
		return;
	}

	/* IP broadcast checks
	*/

	/* leaks are possible here.  we in theory
	 * have the ether header here, and should simply
	 * check for all 1's in the dst mac.  We know we
	 * have an ip address
	 */
	if ((dst.s_addr == 0x0) | (dst.s_addr == 0xffffffff)) {
		fixed_cast.bcast_bytes += caplength;
		return;
	}
	if ((dst.s_addr == ntohl(fixed_cast.ipnet0.s_addr)) ||
		(dst.s_addr == ntohl(fixed_cast.ipnet1.s_addr))) {
		fixed_cast.bcast_bytes += caplength;
		return;
	}

	/* unicast land
	*/
	if (IN_CLASSA(dst.s_addr)) {
		fixed_cast.ucast_bytes += caplength;
		return;
	}
	if (IN_CLASSB(dst.s_addr)) {
		fixed_cast.ucast_bytes += caplength;
		return;
	}
	if (IN_CLASSC(dst.s_addr)) {
		fixed_cast.ucast_bytes += caplength;
		return;
	}
	fixed_cast.xtra_bytes += caplength;
}

/*
 * TBD: put some sort of warning/guard in to warn against too
 * much memory being used (anti - DOS attack).
 *
 * List mechanism here inserts new items at bottom of doubly-linked
 * list and tries to sort them up.  Kept sorted with topN at head of list.
 */
void
topn_ip_filter(unsigned int caplength, struct t_entry *te) 
{
	register struct node *n;
	int rc;
	struct ipkey ik;

	/* ip dst is the key, caplength is the counter
	*/
	/* search and see if ip addr is found
	*/
        ik.ipsrc = te->src.s_addr;
        ik.ipdst = te->dst.s_addr;
        ik.srcport = te->sport;
	ik.dstport = te->dport;
	if (debug) {
		printf("%d %d\n", ik.srcport, ik.dstport);
	}
        ik.ipproto = te->proto;

	n = searchNode(&topn_ip.iplist, &ik, caplength);

	/* failed to find it, insert it
	*/
	if (n == NULL) {
		rc = insertNode(&topn_ip.iplist,  &ik, caplength);
		/* TBD. we need syslog
		*/
		if (rc == 0) {
			fprintf(stderr, "malloc failure - out of memory in topn_ip\n");
		} 
	}
}

/*
 * note: we only print output in the mon.lite file if
 * a filter is configured in
 *
 * don't do something clever - like invoke DNS here
 * this is afterall a wannabe real-time application.
 */
void
write_report(char *monfile, int caught, int dropped, int caught2, int dropped2)
{
	FILE *fd;
	int i;

	fd = fopen(monfile, "w");
	if (fd == NULL ) {
		fprintf(stderr, "oops - can't open monfile %s\n", monfile);
		exit(1);
	}
	fprintf(fd, "pkts: caught:%d : drops:%d: caught2:%d : drop2:%d\n", 
		caught, dropped, caught2, dropped2);

	if (fixed_ipproto.configured) 
	  fprintf(fd, "fixed_ipproto: tcp:%u : udp:%u : icmp:%u : xtra:%u:\n",
			fixed_ipproto.tcp_bytes,
			fixed_ipproto.udp_bytes,
			fixed_ipproto.icmp_bytes,
			fixed_ipproto.xtra_bytes  
			);

	if (fixed_tcp3.configured) 
	  fprintf(fd, "fixed_tcp3: %d:%u : %d:%u : xtra:%u:\n",
			fixed_tcp3.tcp_port1,
			fixed_tcp3.port1_bytes,
			fixed_tcp3.tcp_port2,
			fixed_tcp3.port2_bytes,
			fixed_tcp3.xtra_bytes
			);

	if (fixed_iprange1.configured) {
	  fprintf(fd, "fixed_iprange1: %s-%s:%u : xtra:%u:\n",
			fixed_iprange1.s_ipnet1,
			fixed_iprange1.s_ipnet2,
			fixed_iprange1.range1_bytes,
			fixed_iprange1.extra_bytes
			);
	}

	if (fixed_cast.configured) {
	  fprintf(fd, "fixed_cast: mcast:%u : ucast:%u : bcast:%u : xtra:%u:\n",
			fixed_cast.mcast_bytes,
			fixed_cast.ucast_bytes,
			fixed_cast.bcast_bytes,
			fixed_cast.xtra_bytes 
			);
	}

	if (fixed_size.configured) {
	  fprintf(fd, "fixed_size: tiny:%u : small:%u : med:%u : big:%u:\n",
			fixed_size.tiny,
			fixed_size.small,
			fixed_size.medium,
			fixed_size.large 
			);
	}

	/* printList does all the work.
	 */
	if (topn_ip.configured) {
		printList(fd, &topn_ip.iplist, 1, 1, 1, topn_ip.actualn);
	}

	fclose(fd);
}

void
zero_all_filters()
{
	if (fixed_ipproto.configured) {
		fixed_ipproto.tcp_bytes = fixed_ipproto.udp_bytes = 
		fixed_ipproto.icmp_bytes = fixed_ipproto.xtra_bytes = 0; 
	}

	if (fixed_tcp3.configured) {
		fixed_tcp3.port1_bytes = fixed_tcp3.port2_bytes = 
		fixed_tcp3.xtra_bytes = 0;
	}

	if (fixed_iprange1.configured) {
		fixed_iprange1.range1_bytes = 
		fixed_iprange1.extra_bytes = 0;
	}

	if (fixed_cast.configured) {
		fixed_cast.mcast_bytes = 
		fixed_cast.bcast_bytes = 
		fixed_cast.ucast_bytes = 
		fixed_cast.xtra_bytes = 0;
	}

	if (fixed_size.configured) {
		fixed_size.tiny = 
		fixed_size.small = 
		fixed_size.medium = 
		fixed_size.large = 0;
	}

	/* topN
	*/
	if (topn_ip.configured) {
		destroyList(&topn_ip.iplist);
		initList(&topn_ip.iplist);
	}
}
