/*
 * hashsort.c - deal with dynamic allocation and sorting
 *	of an arbitrary number of nodes.  This is intended
 *	for use by the top n mechanism.
 */

/*
 * Copyright (c) 2001 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "hashsort.h"

#ifndef TRUE
#define TRUE	1
#endif
#ifndef FALSE
#define FALSE	0
#endif

static char *intoa(unsigned int);

/* called before list can be used
*/
void
initList(struct list *l)
{
	l->count = 0;
	l->last = NULL;
	l->total = 0;
	l->tcpcount = l->udpcount = 0;

	/* zero out hash array
	*/
	bzero(&l->hashlist[0], HASHSIZE * sizeof(struct node *));
}

/*
 * searchNode - hash list lookup
 *
 *	inputs:
 *		l - a list pointer
 *		struct ipkey - ipsrc, etc., as logical key
 * 	returns:
 *		NULL - didn't find it.
 *		struct node *n, the node in question.
 *
 * side effect: if node found, increment counter.
 */
struct node *
searchNode(struct list *l, struct ipkey *ipkey, COUNTER counter) 
{
	struct node *n;
	unsigned long ipdst = ipkey->ipdst;
	int mod;

	if (l->count == 0)
		return(NULL);

	if (l->last && (
			l->last->ipkey.ipsrc == ipkey->ipsrc &
			l->last->ipkey.ipdst == ipkey->ipdst &
			l->last->ipkey.srcport == ipkey->srcport &
			l->last->ipkey.dstport == ipkey->dstport &
			l->last->ipkey.ipproto == ipkey->ipproto )) {

		l->last->counter += counter;
		l->total += counter;
		return(l->last);
	}

	/* search the hash/chain list 
	*/
	n = l->hashlist[mod=(ehash((unsigned char *)&ipdst, 4) % HASHSIZE)];

	for (; n != NULL; n = n->hnext) {
		if (n && (
			n->ipkey.ipsrc == ipkey->ipsrc &
			n->ipkey.ipdst == ipkey->ipdst &
			n->ipkey.srcport == ipkey->srcport &
			n->ipkey.dstport == ipkey->dstport &
			n->ipkey.ipproto == ipkey->ipproto )) {

			l->last = n;  
			l->last->counter += counter;
			l->total += counter;
			return(n);
		}
	}
	return(NULL);
}

/*
 *	insertNode
 *
 * 	searchNode succeeds or fails. if it succeeds
 *	we don't call insertNode.  if it fails, we
 *	call this function to create a new hash list
 *	entry and to set the initial counter value.
 *
 *	returns:
 *		1 - ok
 *		0 - malloc failure
 */
int
insertNode (struct list *l, 
	struct ipkey *ipkey, COUNTER counter) 
{
	struct node *prevnode;
	struct node *n;

	n = (struct node *) malloc(sizeof(struct node)); 
	/* TBD. syslog this error.
	*/
	if (n == NULL) {
		fprintf(stderr, "out of space with malloc in insertNode\n");
		return(0);
	}
	/*
	bzero(n, sizeof(struct node));
	*/
	l->count++; 
	if (ipkey->ipproto == IPPROTO_TCP)
		l->tcpcount++;
	else if (ipkey->ipproto == IPPROTO_UDP)
		l->udpcount++;
	n->ipkey = *ipkey; 
	n->counter = counter;
	l->total += counter;
	l->last = n;

	/* insert the node in the hash list
	*/
	{
		struct node *h, *prev;
		unsigned int mod;
		unsigned long ipdst = ipkey->ipdst;

		n->hnext = NULL;
			
		h = l->hashlist[mod=(ehash((unsigned char *)&ipdst, 
			sizeof(unsigned long)) % HASHSIZE)]; 

		if (h == NULL) {
			l->hashlist[mod] = n;
		}
		/* collision, put new node in 1st
		*/
		else {
			n->hnext = h;
			l->hashlist[mod] = n;
		}
	}
	return(1);
}

void
debugList(struct list *l)
{
	dumpHash();
}

/*
 * deallocate malloced hash entries
 */
void
destroyList(struct list *l) 
{
	int i;
	struct node *n;
	struct node *prev;

	for ( i = 0; i < HASHSIZE; i++) {
		n = l->hashlist[i];
		while(n != NULL) {
			prev = n;
			n = n->hnext;
			free(prev);
		}
	}
}

dumpHash(struct list *l)
{
	struct node *h;
	int counters[HASHSIZE];
	int i;
	int ls = 0;

	printf("count %d, total %d\n", l->count, l->total);
	bzero(&counters, sizeof(int) * HASHSIZE);

	printf("\nhash dump--------------\nlist count %d\n", l->count);
	for (i = 0; i < HASHSIZE; i++) {
		h = l->hashlist[i];
		while(h != NULL) {
			ls = 1;
			printf("[%d]: ", i);
			printf("ipsrc %s<->", intoa(h->ipkey.ipsrc));
			printf("ipdst %s ", intoa(h->ipkey.ipdst));
			printf("count %d :", h->counter);
			counters[i]++;  /* total buckets in tier used */
			h = h->hnext;
		}
		if (ls) {
			printf(":[%d]\n", i);
		}
		ls = 0;
	}
	printf("\n");
	for (i = 0; i < HASHSIZE; i++) {
		if (counters[i] != 0) {
			printf("counter[%d]: %d buckets\n", i, counters[i]);
		}
	}
	printf("end hash dump--------------\n");
}

/* hash function
*/
static
unsigned long 
ehash (unsigned char *name, int size)
{
	unsigned long h = 0, g;

	while (size--) {
		h = (h << 4) + *name++;
		if ( g = h & 0xF0000000 )
			h ^= g >> 24;
		h &= ~g; 
	}
	return(h);
}

/*
 * ok, we have a hash list.  pack it into an array
 * and return the array.
 *
 *	returns:
 *		1 - packed array exists
 *		0 - malloc failure
 */
static
struct node *
packArray(struct list *l)
{
	struct node *head, *n;
	unsigned int c = l->count;	/* count of hash entries allocated */
	int i = 0;
	int hcounter=0;

	/* create the "array"
	*/
	head = (struct node *) malloc(c * sizeof (struct node));
	/* TBD. syslog this
	*/
	if (head == NULL) {
		fprintf(stderr, "packArray, malloc failed\n");
		return(NULL);
	}

	/* walk the hashlist
	*/
	for ( i = 0; i < HASHSIZE; i++) {
		n = l->hashlist[i];
		/* if n is non-null, we have a node
		 * in use, therefore put it in the array,
		 * and increment hcounter
		 */
		while (n != NULL) {
			head[hcounter].ipkey = n->ipkey; 
			head[hcounter].counter = n->counter;
			/* walk to end of linked list repeating previous
			 * operation
			 */
			n = n->hnext;
			hcounter++;
		}
	}
	return(head);
}

static int      
sortbycounter(e1, e2)
struct node *e1, *e2;
{
        if (e1->counter > e2->counter) return -1;
        if (e1->counter < e2->counter) return 1;
        return 0;
}       



/*
 * borrowed from trafshow.
 *
 * A faster replacement for inet_ntoa().
 */

static
char *
intoa(addr)
unsigned int addr;
{
	register char *cp;
	register unsigned int byte;
	register int n;
	static char buf[sizeof(".xxx.xxx.xxx.xxx")];

	addr = ntohl(addr);
	cp = &buf[sizeof buf];
	*--cp = '\0';

	n = 4;
	do {
		byte = addr & 0xff;
		*--cp = byte % 10 + '0';
		byte /= 10;
		if (byte > 0) {
			*--cp = byte % 10 + '0';
			byte /= 10;
			if (byte > 0)
				*--cp = byte + '0';
		}
		*--cp = '.';
		addr >>= 8;
	} while (--n > 0);

	return cp + 1;
}

static
char *
prototostring(int p)
{
	if (p == IPPROTO_TCP)
		return("tcp");
	else if (p == IPPROTO_UDP)
		return("udp");
	else if (p == IPPROTO_ICMP)
		return("icmp");
	return (NULL);
}

/*
 * given a list, sort it,
 * and print topN results.
 * if tcp flag, print top N tcp.
 * if udp flag, print top N udp,
 *
 * inputs:
 *	fd - file for print output
 *	l - hash array
 *	anyflag/tcpflag/udpflags - flags to indicates various filter outputs
 *	count - max desired outputs for all filters
 */
printList(FILE *fd, struct list *l, int anyflag, int tcpflag, int udpflag, int count) 
{
	struct node *head;
	char *ps;
	char buf[1024];
	char b1[1024], b2[1024];
	int i;
	unsigned long subtotal = 0;
	unsigned long total = l->total;
	int protohasports = 0;
	int tcpcount, udpcount;

	int c = l->count;	/* no of hash elements */

	tcpcount = udpcount = 0;

	/* form the hashlist into a packed array
	*/
	head = packArray(l);
	if (head == NULL) {
		fprintf(fd, "malloc error in printList\n");
		return;
	}

	/* quicksort it
	*/
	qsort((void *)head, c, sizeof (struct node), sortbycounter);

	/* print it out
	*/
	if (anyflag) {
		/* walk the array for anyflag
		*/
		fprintf(fd, "topn_ip : %u : ", c + 1); /* xtra is counted in */ 
		for ( i = 0; i < c; i++) {
			if (i >= count) {
				break;
			}
			subtotal += head[i].counter;
			/* produce a string for the proto type
			 * if possible, else a number.
			 */
			ps = prototostring(head[i].ipkey.ipproto);
			if (ps == NULL) {
				sprintf(buf, "%d", head[i].ipkey.ipproto);
				ps = &buf[0];
			}
			else {
				protohasports = 1;
			}
			/* intoa can only do one ip address at a time,
			 * therefore we must buffer one
			 */
			sprintf(b1, "%s", intoa(head[i].ipkey.ipdst));
			/* if we have non-zero ports include them,
			 * else drop them.
			 */
			if (protohasports) {
				fprintf(fd,"%s.%d->%s.%d(%s): %u : ",
					intoa(head[i].ipkey.ipsrc),
					head[i].ipkey.srcport,
					b1,
					head[i].ipkey.dstport,
					ps,
					head[i].counter);
			}
			else {
				fprintf(fd,"%s->%s(%s):%u : ",
					intoa(head[i].ipkey.ipsrc),
					b1,
					ps,
					head[i].counter);
			}
		}
		fprintf(fd, " xtra: %u\n", total - subtotal);
	}
	if (tcpflag) {
		fprintf(fd, "topn_tcp : %u : ", l->tcpcount); 
		for ( i = 0; i < c; i++) {
			if (head[i].ipkey.ipproto != IPPROTO_TCP)
				continue;
			tcpcount++;
			if (tcpcount > count) {
				break;
			}
			/*
			subtotal += head[i].counter;
			*/
			sprintf(b1, "%s", intoa(head[i].ipkey.ipdst));
			fprintf(fd, "%s.%d->%s.%d: %u : ",
				intoa(head[i].ipkey.ipsrc),
				head[i].ipkey.srcport,
				b1,
				head[i].ipkey.dstport,
				head[i].counter);
		}
		/*
		fprintf(fd, " xtra: %u\n", total - subtotal);
		*/
		fprintf(fd, "\n");
	}
	if (udpflag) {
		fprintf(fd, "topn_udp : %u : ", l->udpcount); 
		for ( i = 0; i < c; i++) {
			if (head[i].ipkey.ipproto != IPPROTO_UDP) {
				continue;
			}
			udpcount++;
			if (udpcount > count) {
				break;
			}
			/*
			subtotal += head[i].counter;
			*/
			sprintf(b1, "%s", intoa(head[i].ipkey.ipdst));
			fprintf(fd, "%s.%d->%s.%d: %u : ",
				intoa(head[i].ipkey.ipsrc),
				head[i].ipkey.srcport,
				b1,
				head[i].ipkey.dstport,
				head[i].counter);
		}
		/*
		fprintf(fd, " xtra: %u\n", total - subtotal);
		*/
		fprintf(fd, "\n");
	}

	free(head);

}

#ifdef TESTMAIN

static int f = 0;

unsigned long
getRandomWord()
{
	static unsigned long rw;

#define REALLYRANDOM
#ifdef REALLYRANDOM
	if (f == 0) {
		f = open("/dev/urandom", 0);
		if (f < 0) {
			perror("open");
			exit(1);
		}
	}
	read(f, &rw, sizeof(unsigned long));
#else
	rw = rand();
#endif
	return(rw);
}

functionTest()
{
	struct list l;
	COUNTER c;
	struct ipkey ik;
	int rc;
	struct node *n = NULL;

	/* load sample ipkey and counter
	*/
	ik.ipsrc = inet_addr("1.1.1.1");
	ik.ipdst = inet_addr("1.1.1.2");
	ik.srcport = ik.dstport = 0;
	ik.ipproto = IPPROTO_TCP;
	c = 1;	

	initList(&l);

	n = searchNode(&l, &ik, c);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 
	
	rc = insertNode(&l, &ik, c);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	n = searchNode(&l, &ik, c);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 

	/* add 4 more nodes and dump the list
	*/
	ik.ipsrc = inet_addr("1.1.1.2");
	ik.ipdst = inet_addr("1.1.1.3");
	c = 2;
	rc = insertNode(&l, &ik, c);
	if (rc == 0) {
		printf("insertNode failed\n");
	}
	ik.ipsrc = inet_addr("1.1.1.3");
	ik.ipdst = inet_addr("1.1.1.4");
	c = 3;
	ik.ipproto = IPPROTO_UDP;
	rc = insertNode(&l, &ik, c);
	if (rc == 0) {
		printf("insertNode failed\n");
	}
	ik.ipsrc = inet_addr("1.1.1.4");
	ik.ipdst = inet_addr("1.1.1.5");
	ik.ipproto = 1;
	c = 4;
	rc = insertNode(&l, &ik, c);
	if (rc == 0) {
		printf("insertNode failed\n");
	}
	ik.ipsrc = inet_addr("1.1.1.6");
	ik.ipdst = inet_addr("1.1.1.7");
	c = 5;
	rc = insertNode(&l, &ik, c);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	/* now search for a given node, and increment its
	 * value
	 */
	c = 5;
	n = searchNode(&l, &ik, c);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 
	else {
		printf("searchNode found node\n");
	}

	/* now search for a previous node, note last
	 * as a simple cache should not be used here
	 */
	ik.ipsrc = inet_addr("1.1.1.4");
	ik.ipdst = inet_addr("1.1.1.5");
	c = 1;
	n = searchNode(&l, &ik, c);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 
	else {
		printf("searchNode found node\n");
	}

	dumpHash(&l);

	printList(STDOUT, &l, 1, 1, 1, 5);

	destroyList(&l);
}

#ifdef LATER
randomTest()
{
	struct list l;
	unsigned long key1;
	COUNTER counter1;
	unsigned long k,c;
	struct node *n;
	int i;
	int rc;

	initList(&l);
	
	for ( i = 0; i < 10000; i++) {
		key1 = getRandomWord();
		counter1 = getRandomWord();

		printf("[i=%d], key=%lx, counter=%lx\n", i, key1, counter1);

		n = searchNode(&l, key1);
		/* did not find it
		*/
		if ( n == NULL) {
			rc = sortNode(&l,  NULL, FALSE, key1, counter1);
			if (rc == 0) {
				printf("sortNode failed\n");
				exit(1);
			}
		}
		/* did find it
		*/
		else {
			printf("FOUND ONE!: [i=%d], key=%lx", i, key1);
			rc = sortNode(&l,  n, TRUE, key1, counter1);
			if (rc == 0) {
				printf("sortNode failed\n");
				exit(1);
			}
		}
	}
	
	printf("node count %d\n", l.count);

	initWalk(&l);

	i = 0;
	while( rc = walkList(&l, &k, &c)) {
		printf("[%d], tuple: (k=%lx, c=%lx)\n", i++, k, c);
	} 

	dumpHash(&l);

	destroyList(&l);
}

#endif

main()
{
	functionTest();
}
#endif

