/*
 * config.c - config info for ourmon utility
 * 
 * purpose: read/parse/store tag values from /etc/ourmon.conf configuration
 * file.
 * 
 * file syntax: ascii lines, one tag per line, may have N values depending on
 * tag skip lines/and comments, # in 1st column is comment
 *
 * All filters are configured with one line, the name of the filter
 * comes first.  optional parameters may follow on a per filter basis.
 * Filter spec is provided below.
 * 
 * tag syntax:
 * 
 * 0. pkts  
 *
 *	There is an implicit filter output that counts 	packets caught 
 *	and dropped.  These variables are zeroed
 *	at this point by the less than wonderful idea of
 *	zeroing out the BPF at write output file time.  
 *      The information comes from pcap (pcap_stats()).
 *	It allows us to determine if the filter/s are causing
 *	packet drops.  Note that two interfaces can be supported.
 *
 * 1. fixed_ip
 * 
 *	syntax: fixed_ip
 *
 *      Fixed filter; i.e., keys are not dynamic/changeable.
 *	tcp/udp/icmp/remainder (4 keys).
 *	This filter is zeroed out at every write to output.
 *
 * 2. fixed_tcp3
 *
 *	syntax:  fixed_tcp3 tcp_port1  tcp_port2
 *
 *	Fixed filter.  captures and counts packets with tcp dest or
 *	src port equal to port1 or port2.  counts packets (extra)
 *	that do not belong to that filter spec of course.
 *	Note tcp_port1 for example should be a well-known port
 *	and will cause the count of packets with src OR dest port
 *	set to that value. This filter is zeroed out at every write 
 *	to output.
 *
 * 3. fixed_iprange1
 *
 *	syntax: iprange1  ipaddr1 ipnetmask1  ipaddr2 ipnetmask2 s1 s2
 *
 *	s1 and s2 are strings that must be supplied as they are used
 *	in the output. s1 should describe the 1st network, and s2
 *	should describe the 2nd network.
 *
 *	Measures total ip exchange in bytes between first ip network
 *	and second ip network.  Remainder to extra counter; i.e.,
 *	two numbers will be printed out.  More ipranges may be added
 *	later.  This is the semantic equal to tcpdump:
 *		net X and net Y
 *
 *  4. spec_ipproto4 - TBD
 *	
 *	syntax: spec_ipproto4 proto1 proto2 proto3
 *
 *	Three ip protos may be specified (e.g., OSPF, AH, ESP)
 *	in order to check on explicit IP level protocols.  A remainder
 *	bucket exists (xtra).  This filter is zeroed out per alarm.
 *	Of course the fixed protocol example could be done here as well.
 *	E.g., tcp, udp, icmp.  
 *
 *  5. fixed_cast 
 *	
 *	syntax:   fixed_cast  ip_addr  netmask
 *
 *	The focus of this filter is on dividing up packets
 *	into broadcast/multicast/unicast groups.  If the
 *	interface/s used are ethernet interfaces, then the
 *	filter bases its decisions on ethernet dst addresses.
 *	If the interface/s used are NOT ethernet interfaces,
 *	then IP addresses are used.  An IP address (which is used
 *	for guessing at the broadcast IP address) must be supplied.
 *	There are 4 buckets, multicast, unicast, broadcast
 *	and xtra.  Note that if we are running on ethernet,
 *	then decisions are 100% based on ethernet level information,
 *	and the results are not IP specific.  If non-ethernet interfaces
 *	are used,  the xtra bucket will capture all other traffic, which could
 *	include ARP, non-IP, and some directed broadcast, missed because
 *	the IP/netmask info above may miss some kinds of directed broadcast
 *	in a e.g., VLSM-based network.
 *
 * 6. fixed_size:  
 *
 *	syntax: fixed_size
 *
 *	4 buckets for measuring packet sizes (at layer 2):
 *	essentially tiny, small, medium, mtu-ish (1500)
 *
 *	tiny: <= 100
 *	small: < 500
 *	medium: < 1000
 *	large: <= 1500
 *
 *	Currently cannot reset bucket sizes.
 *
 * ------- above are static, remainder are dynamic (topn) -----------
 *
 *  7. topn_ip: 
 *	syntax: topn_ip N
 *
 *	N is the desired # of top N ip destinations to be output.
 *	N must be less than 10 currently.  The output will consist
 *	of a sorted pair of (ip dst, bytes) for the time period.
 *	Note that at alarm time, the last set of sample buckets are
 *	reinitialized (we start over).  TBD: There is a rough internal
 *	limit of 64k samples.
 *
 *    Now includes udp and tcp graphs as well (3 in one).
 *
 */

/*
 * Copyright (c) 2001 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>

#include "./filter.h"

#define MAXLINE	1024

/* global filter variables *************************** */
struct fixed_ipproto fixed_ipproto;
struct fixed_tcp3 fixed_tcp3;
struct fixed_iprange1 fixed_iprange1;
struct spec_ipproto spec_ipproto;
struct fixed_cast fixed_cast;
struct fixed_size fixed_size;

int guard1 = 1;
struct topn_ip topn_ip;
int guard2 = 2;

/*************************************************/

extern int debug;

/*
 * function prototype declarations
 */
static          switchOnOp();
int             resolveAddress(char *addr, struct sockaddr_in * to);

void
configInit()
{
	fixed_ipproto.configured = FALSE;
	fixed_tcp3.configured = FALSE;
	fixed_iprange1.configured = FALSE;
	fixed_cast.configured = FALSE;
	fixed_size.configured = FALSE;
	topn_ip.configured = FALSE;
}

/*
 * read the config file and load its values
 * 
 * return: 0: ok -1: fatal error (although we return)
 */
int
configRead(configName)
	char           *configName;	/* name of config file */
{
	FILE           *fd;
	char            linebuf[1024];
	char            opbuf[1024];
	char           *opword;
	char           *result;
	int             lineNumber = 1;
	int 		i, j;
	char 		*ptr;
	int 		ccount;

	fd = fopen(configName, "r");
	if (fd == NULL) {
		fprintf(stderr, "ourmon: open of config file %s failed\n",
		       configName);
		exit(1);
	}

	configInit();

	/*
	 * while not at EOF read a line, parse it
	 */
	while (feof(fd) == 0) {
		result = fgets(linebuf, MAXLINE, fd);
		if (result == NULL)
			break;

		ccount = strlen(linebuf);
		linebuf[ccount-1] = 0;

		/*
		 * skip comment
		 */
		if (linebuf[0] == '#') {
			lineNumber++;
			continue;
		}
		/*
		 * skip empty lines
		 */
		if (linebuf[0] == '\n' || linebuf[0] == '\r') {
			lineNumber++;
			continue;
		}
		/*
		 * skip whitespace in line
		 */
		while (*result == ' ' || *result == '\t') {
			result++;
		}
		/*
		 * find first token
		 */
		strncpy(opbuf, result, 1024);
		opword = strtok(opbuf, " \t");
		if (opword == NULL || (strlen(opword) == 0))
			continue;
		/*
		 * sscanf(result, "%s %s", op, value); printf("op value %s
		 * %s\n", op, value);
		 */
		switchOnOp(opword, result, lineNumber);
		lineNumber++;
	};
	fclose(fd);

	if (debug) {
		configDump();
	}

	/*
	 * semantic checks if any
	 */
	return (0);

}  /* end configRead() */

static
switchOnOp(op, linebuf, line)
	char           *op;
	char           *linebuf;
	int             line;
{
	char            dc[1024];
	char            v1[1024];
	char            v2[1024];
	char            v3[1024];
	char            v4[1024];
	char            v5[1024];
	char            v6[1024];
	int             rc;
	struct sockaddr_in sa;

	/* mn ifname ip */
	if (strcmp(op, "fixed_ipproto") == 0) {
		fixed_ipproto.configured = TRUE;
	} else if (strcmp(op, "fixed_tcp3") == 0) {
		fixed_tcp3.configured = TRUE;
		sscanf(linebuf, "%s %s %s", dc, v1, v2);
		fixed_tcp3.tcp_port1 = atoi(v1);
		fixed_tcp3.tcp_port2 = atoi(v2);
		if (fixed_tcp3.tcp_port1 == 0 || fixed_tcp3.tcp_port2 == 0) {
			fprintf(stderr, "fixed_tcp3 needs two non-zero port arguments\n");
			fprintf(stderr, "syntax: fixed_tcp3 port1 port2\n");
			exit(1);
		}
	} else if (strcmp(op, "fixed_iprange1") == 0) {
		fixed_iprange1.configured = TRUE;

		sscanf(linebuf, "%s %s %s %s %s %s %s", dc, v1, v2, v3, v4, v5, v6);
		/* labels first
		*/
		strncpy(fixed_iprange1.s_ipnet1, v5, 1024);
		strncpy(fixed_iprange1.s_ipnet2, v6, 1024);

		/* now 4 "ip" addresses, two of which should actually
		 * be subnet masks
		 */
		rc = resolveAddress(v1, &sa);

		if (rc <= 0) {
			fprintf(stderr, "fixed_iprange1: invalid ip address 1 %s\n", v1);
			exit(1);
		}
		/* in network byte order
		*/
		fixed_iprange1.ipaddr1.s_addr = sa.sin_addr.s_addr;

		rc = resolveAddress(v2, &sa);
		if (rc <= 0) {
			fprintf(stderr, "fixed_iprange1: invalid mask 1 %s\n", v2);
			exit(1);
		}
		/* in network byte order
		*/
		fixed_iprange1.netmask1.s_addr = sa.sin_addr.s_addr;

		rc = resolveAddress(v3, &sa);
		if (rc <= 0) {
			fprintf(stderr, "fixed_iprange1: invalid ip address 2 %s\n", v1);
			exit(1);
		}
		/* in network byte order
		*/
		fixed_iprange1.ipaddr2.s_addr = sa.sin_addr.s_addr;

		rc = resolveAddress(v4, &sa);
		if (rc <= 0) {
			fprintf(stderr, "fixed_iprange1: invalid netmask 2 %s\n", v1);
			exit(1);
		}
		fixed_iprange1.netmask2.s_addr = sa.sin_addr.s_addr;

		/* in network byte order
		*/

		fixed_iprange1.ipnet1.s_addr = fixed_iprange1.ipaddr1.s_addr
					& fixed_iprange1.netmask1.s_addr;
		fixed_iprange1.ipnet2.s_addr = fixed_iprange1.ipaddr2.s_addr
					& fixed_iprange1.netmask2.s_addr;

	} else if (strcmp(op, "fixed_cast") == 0) {
		fixed_cast.configured = TRUE;

		sscanf(linebuf, "%s %s %s", dc, v1, v2);

		/* now 2 "ip" addresses, ip addr and net mask
		 */
		rc = resolveAddress(v1, &sa);

		if (rc <= 0) {
			fprintf(stderr, "fixed_cast: invalid ip address %s\n", v1);
			exit(1);
		}
		/* in network byte order
		*/
		fixed_cast.ipaddr1.s_addr = sa.sin_addr.s_addr;

		rc = resolveAddress(v2, &sa);
		if (rc <= 0) {
			fprintf(stderr, "fixed_cast: invalid mask %s\n", v2);
			exit(1);
		}
		/* in network byte order
		*/
		fixed_cast.netmask1.s_addr = sa.sin_addr.s_addr;

		/* now compute all 0's and all 1's bcast
		*/

		fixed_cast.ipnet0.s_addr = fixed_cast.ipaddr1.s_addr
					& fixed_cast.netmask1.s_addr;

		fixed_cast.ipnet1.s_addr = fixed_cast.ipnet0.s_addr
					| ~fixed_cast.netmask1.s_addr;

	} else if (strcmp(op, "fixed_size") == 0) {
		fixed_size.configured = TRUE;
	} else if (strcmp(op, "topn_ip") == 0) {
		topn_ip.configured = TRUE;
		sscanf(linebuf, "%s %s", dc, v1);
		topn_ip.actualn = atoi(v1);
		if (topn_ip.actualn > MAXTOPN) {
			fprintf(stderr, "topn_ip topN limit is %d\n",
				MAXTOPN);
			exit(1);
		}
		/* initialize the list mechanism
		*/
		initList(&topn_ip.iplist);
	} else {
		fprintf(stderr, "switchOnOp(): line [%d], unknown config op (%s) in config file\n", line, op);
		exit(1);
	}
}

/*
 * debug - dump configuration variables
 */
configDump()
{
	int             i;

	printf("---------config dump-------------------\n");
	if (fixed_ipproto.configured) {
          printf("fixed_ipproto.configured %d\n", fixed_ipproto.configured);
	}
	if (fixed_tcp3.configured) {
		printf("fixed_tcp3: port1 %d\n", fixed_tcp3.tcp_port1);
		printf("fixed_tcp3: port2 %d\n", fixed_tcp3.tcp_port2);
	}
	if (fixed_iprange1.configured) {
		printf("iprange1: ipaddr1 %s, ", 
			inet_ntoa((struct in_addr)fixed_iprange1.ipaddr1));
		printf("netmask1 %s\n", 
			inet_ntoa((struct in_addr)fixed_iprange1.netmask1) 
			);
		printf("ipaddr2 %s, ", 
			inet_ntoa((struct in_addr)fixed_iprange1.ipaddr2));
		printf("netmask2 %s\n", 
			inet_ntoa((struct in_addr)fixed_iprange1.netmask2) 
			);
		printf("ipnet1 %s, ", 
			inet_ntoa((struct in_addr)fixed_iprange1.ipnet1));
		printf("ipnet2 %s\n", 
			inet_ntoa((struct in_addr)fixed_iprange1.ipnet2));
		printf("labels %s %s\n",
			fixed_iprange1.s_ipnet1,
			fixed_iprange1.s_ipnet2);
	}
	if (topn_ip.configured) {
		printf("topn_ip: actualn %d\n", topn_ip.actualn);
	}

	if (fixed_cast.configured) {
		printf("fixed_cast: bcast 1's %s \n", 
			inet_ntoa((struct in_addr)fixed_cast.ipnet1));
		printf("fixed_cast: bcast 0's %s \n", 
			inet_ntoa((struct in_addr)fixed_cast.ipnet0));
	}
	if (fixed_size.configured) {
		printf("fixed_size: ON \n");
	}
	printf("---------config dump-------------------\n");
}

#ifdef TESTMAIN
int debug = 1;

char *program_name;

main(argc, argv)
int argc; 
char **argv;
{

	configRead(argv[1]);
	/* implicit
	configDump();
	*/

}
#endif

